"""Tests for the ``MPool`` class"""

from time import sleep
from unittest import TestCase

from egon.nodes import MPool


def target_func() -> None:
    """A dummy target function"""

    sleep(2)


class ProcessAllocation(TestCase):
    """Test instances fork the correct number of processes"""

    def setUp(self) -> None:
        self.num_processes = 4
        self.pool = MPool(self.num_processes, target_func)

    def test_allocation_at_init(self) -> None:
        """Test the correct number of processes are allocated at init"""

        self.assertEqual(self.num_processes, self.pool.num_processes)
        self.assertEqual(self.num_processes, len(self.pool._processes))

    def test_error_on_negative_processes(self) -> None:
        """Assert a value error is raised when the ``num_processes`` attribute is set to a negative"""

        with self.assertRaises(ValueError):
            MPool(-1, target_func)


class Execution(TestCase):
    """Tests for the starting, running, and stopping of allocated processes"""

    def test_pool_is_finished_after_execution(self) -> None:
        """Test the ``is_finished`` property is updated after the pool executes"""

        pool = MPool(2, target_func)
        self.assertFalse(pool.is_finished(), 'Default finished state is not False.')

        pool.start()
        pool.join()

        self.assertTrue(pool.is_finished())

    def test_processes_killed_on_command(self) -> None:
        """Test processes are killed on demand"""

        pool = MPool(1, lambda *args: sleep(10))
        pool.start()
        pool.kill()
        pool.join()

        self.assertFalse(pool.is_finished())
        self.assertTrue(
            pool._processes[0].exitcode < 0,
            f'Process not ended by termination signal ({pool._processes[0].exitcode})')


class ZeroProcessPool(TestCase):
    """Tests for pool behavior with zero processes"""

    @classmethod
    def setUpClass(cls) -> None:
        cls.pool = MPool(0, target_func)

    def test_pool_is_finished_by_default(self) -> None:
        """Test the ``is_finished`` is true for a pool with zero processes"""

        self.assertTrue(MPool(0, target_func).is_finished)

    def test_run_error(self):
        """Test running the pool with zero processes raises an error"""

        with self.assertRaises(RuntimeError):
            self.pool.start()

    def test_join_error(self):
        """Test joining the pool with zero processes raises an error"""

        with self.assertRaises(RuntimeError):
            self.pool.join()

    def test_kill_error(self):
        """Test killing the pool with zero processes raises an error"""

        with self.assertRaises(RuntimeError):
            self.pool.kill()



"""Tests for the ``BaseConnector`` class"""

from unittest import TestCase

from egon.connectors import BaseConnector


class DefaultState(TestCase):
    """Test connector properties are null by default"""

    def test_default_parent_is_none(self) -> None:
        """Test that a connector's parent is ``None`` by default"""

        self.assertIsNone(BaseConnector().parent_node)

    def test_default_partners_is_empty(self) -> None:
        """Test the default tuple of partners is empty"""

        self.assertIsInstance(BaseConnector().partners, tuple)
        self.assertFalse(BaseConnector().partners)

    def test_default_connection_false(self) -> None:
        """Test connectors are marked as disconnected by default"""

        self.assertFalse(BaseConnector().is_connected())



"""Dash components that provide preconfigured plots for visualizing data."""

from __future__ import annotations

from datetime import datetime
from typing import TYPE_CHECKING

import dash_core_components as dcc
import plotly.express as px
import psutil

if TYPE_CHECKING:  # pragma: no cover
    from egon.pipeline import Pipeline


class CustomGraph(dcc.Graph):
    """The same as a Dash Graph instance but with modified default values

    Modified Defaults
        animate: True
        displayModeBar: False
    """

    def __init__(self, *args, **kwargs) -> None:
        kwargs.setdefault('animate', True)
        kwargs.setdefault('config', dict())
        kwargs['config'].setdefault('displayModeBar', False)
        super().__init__(*args, **kwargs)


class CpuPercentageGraph(CustomGraph):
    """A graph representing the host system's CPU usage over time"""

    def __init__(self, *args, **kwargs) -> None:
        """Plot the percentage usage for each CPU core over time

        Args:
            Any arguments for a dash ``Graph`` object except ``figure``
        """

        cpu_labels = []
        plot_data = {'Time': [datetime.now()]}
        for i in range(psutil.cpu_count()):
            label = f'CPU{i}'
            cpu_labels.append(label)
            plot_data[label] = [0]

        fig = px.line(plot_data, x='Time', y=cpu_labels, labels={'value': 'CPU %'})
        fig.layout.update({'yaxis': {'range': [0, 100]}})
        super().__init__(*args, **kwargs, figure=fig)


class RamPercentageGraph(CustomGraph):
    """A graph representing the host system's RAM usage over time"""

    def __init__(self, *args, **kwargs) -> None:
        """Plot the percentage of RAM used over time

        Args:
            Any arguments for a dash ``Graph`` object except ``figure``
        """

        mem_usage_at_init = psutil.virtual_memory()[2]

        kwargs.setdefault('animate', True)
        fig = px.line({'Time': [datetime.now()], 'Memory (%)': [mem_usage_at_init]}, x='Time', y='Memory (%)')
        fig.layout.update({'yaxis': {'range': [0, 100]}})
        super().__init__(*args, **kwargs, figure=fig)


class PipelineQueueSize(CustomGraph):
    """A graph representing the number of queued objects in a pipeline"""

    def __init__(self, pipeline: Pipeline, *args, **kwargs) -> None:
        """Plot the size of a pipeline's input queues over time

        Args:
            Any arguments for a dash ``Graph`` object except ``figure``
        """

        node_labels = []
        plot_data = {'Time': [datetime.now()]}
        for input_connector in pipeline.connectors[0]:
            node_labels.append(input_connector.name)
            plot_data[input_connector.name] = [0]

        fig = px.line(plot_data, x='Time', y=node_labels, labels={'value': 'Queue Size'})
        super().__init__(*args, **kwargs, figure=fig)



"""Dash components that provide a cytoscape-like representation of running
pipelines.
"""

from __future__ import annotations

from itertools import chain
from pathlib import Path
from typing import List, TYPE_CHECKING

import dash_cytoscape as cyto
import yaml

from egon.connectors import Input, Output
from egon.nodes import AbstractNode, Node, Source

if TYPE_CHECKING:  # pragma: no cover
    from egon.pipeline import Pipeline

STYLE_PATH = Path(__file__).resolve().parent.parent / 'assets' / 'cyto_style.yml'


class PipelineCytoscape(cyto.Cytoscape):
    """A Dash compatible component for visualizing pipelines with cytoscape"""

    def __init__(self, pipeline: Pipeline, **kwargs) -> None:
        """An interactive, cytoscape style plot of a constructed pipeline

        If the ``stylesheet`` argument is provided, the default style sheet for
        this class is overwritten.

        Args:
            pipeline: The pipeline that will be visualized
            kwargs: Any additional ``Cytoscape`` arguments except for ``elements``
        """

        stylesheet = kwargs.pop('stylesheet', self.default_stylesheet())

        # Set customized defaults
        kwargs.setdefault('minZoom', 0.25)
        kwargs.setdefault('maxZoom', 2)
        kwargs.setdefault('style', dict())
        kwargs['style'].setdefault('width', '100%')
        kwargs['style'].setdefault('height', '800px')

        elements = []
        for node in chain(*pipeline.nodes):
            # Add the pipeline node to the figure
            node_id = str(id(node))
            elements.append({
                'data': {'id': node_id, 'label': node.name},
                'classes': self._get_classes(node)
            })

            for connector in chain(*node.connectors):
                # Add each connector to the figure
                connector_id = str(id(connector))
                elements.append({
                    'data': {'id': connector_id, 'parent': node_id, 'label': connector.name},
                    'classes': self._get_classes(connector)
                })

                # Draw an arrow between the connector and each of its partners
                if isinstance(connector, Input):
                    for partner in connector.partners:
                        partner_id = str(id(partner))
                        elements.append({
                            'data': {
                                'source': partner_id,
                                'source_label': partner.name,
                                'target': connector_id,
                                'target_label': connector.name
                            }
                        })

        super().__init__(elements=elements, stylesheet=stylesheet, **kwargs)

    @staticmethod
    def default_stylesheet() -> List[dict]:
        """Return a copy of the default style sheet

        Return:
            A list of style settings
        """

        with STYLE_PATH.open() as infile:
            return yaml.safe_load(infile)

    @staticmethod
    def _get_classes(node: AbstractNode) -> str:
        """Return the CSS class of a plotted node

        Return value depends on the type of node (e.g., source or target)

        Args:
            node: The node to return the class for

        Returns:
            The CSS class of the node
        """

        if isinstance(node, Input):
            return 'connector input'

        if isinstance(node, Output):
            return 'connector output'

        if isinstance(node, Node):
            return 'pipeline-node inline'

        if isinstance(node, Source):
            return 'pipeline-node source'

        return 'pipeline-node target'



"""Tests for the ``ObjectCollection`` class."""

from unittest import TestCase

from egon.utils import ObjectCollection


class Add(TestCase):
    """Test the addition of data to the collection"""

    def test_values_are_added(self) -> None:
        """Test the ``add`` method adds values to the collection"""

        collection = ObjectCollection()
        collection.add(1)
        self.assertIn(1, collection)

    def test_no_duplicate_entries(self) -> None:
        """Test the collection does not store duplicate entries"""

        collection = ObjectCollection()
        collection.add(1)
        collection.add(1)
        self.assertEqual(1, len(collection))


class Remove(TestCase):
    """Test the removal of data from the collection"""

    def runTest(self) -> None:
        collection = ObjectCollection()
        collection.add(1)
        collection.remove(1)
        self.assertNotIn(1, collection)


class CastList(TestCase):
    """Test the casting of a collection to a list"""

    def runTest(self) -> None:
        test_data = [1, 2, 3, 4]
        self.assertListEqual(list(ObjectCollection(test_data)), test_data)



"""Generic utility classes"""

from typing import Hashable, Collection, Iterable, Optional


class KillSignal:
    """Used to indicate that a process should exit"""


class ObjectCollection:
    """Collection of objects with O(1) add and remove"""

    def __init__(self, data: Optional[Collection] = None) -> None:
        """A mutable collection of arbitrary objects

        Args:
            data: Populate the collection instance with the given data
        """

        # Map object hash values to their index in a list
        self._object_list = list(set(data)) if data else []
        self._index_map = {o: i for i, o in enumerate(self._object_list)}

    def add(self, obj: Hashable) -> None:
        """Add a hashable object to the collection

        Args:
            obj: The object to add
        """

        # Exit if ``x`` is already in the collection
        if obj in self._index_map:
            return

        # Add ``x`` to the end of the collection
        self._index_map[obj] = len(self._object_list)
        self._object_list.append(obj)

    def remove(self, obj: Hashable) -> None:
        """Remove an object from the collection

        Args:
            obj: The object to remove
        """

        index = self._index_map[obj]

        # Swap element with last element so that removal from the list can be done in O(1) time
        size = len(self._object_list)
        last = self._object_list[size - 1]
        self._object_list[index], self._object_list[size - 1] = self._object_list[size - 1], self._object_list[index]

        # Update hash table for new index of last element
        self._index_map[last] = index

        del self._index_map[obj]
        del self._object_list[-1]

    def __iter__(self) -> Iterable:
        return iter(self._object_list)

    def __contains__(self, item: Hashable) -> bool:
        return item in self._object_list

    def __len__(self) -> int:
        return len(self._object_list)

    def __str__(self) -> str:  # pragma: no cover
        return f'<Container({self._object_list})>'



"""Tests for ``Input`` connector objects."""

import time
from unittest import TestCase

from egon.connectors import Input, KillSignal
from egon.exceptions import MissingConnectionError
from egon.mock import MockTarget


class InputGet(TestCase):
    """Test data retrieval from ``Input`` instances"""

    def setUp(self) -> None:
        self.input_connector = Input()

    def test_error_on_non_positive_refresh(self) -> None:
        """Test a ValueError is raised when ``refresh_interval`` is not a positive number"""

        with self.assertRaises(ValueError):
            self.input_connector.get(timeout=15, refresh_interval=0)

        with self.assertRaises(ValueError):
            self.input_connector.get(timeout=15, refresh_interval=-1)

    def test_returns_queue_value(self) -> None:
        """Test the ``get`` method retrieves data from the underlying queue"""

        test_val = 'test_val'
        self.input_connector._queue.put(test_val)

        time.sleep(1)
        self.assertEqual(test_val, self.input_connector.get(timeout=1000))

    def test_timeout_raises_timeout_error(self) -> None:
        """Test a ``TimeoutError`` is raise on timeout"""

        with self.assertRaises(TimeoutError):
            self.input_connector.get(timeout=1)

    def test_kill_signal_on_finished_parent_node(self) -> None:
        """Test a kill signal is returned if the parent node is finished"""

        target = MockTarget()
        self.assertFalse(target.is_expecting_data())
        self.assertIs(target.input.get(timeout=15), KillSignal)


class InputIterGet(TestCase):
    """Test iteration behavior of the ``iter_get`` method"""

    def setUp(self) -> None:
        """Create an input connector that is assigned to a parent node

        The input connector must be assigned to a parent node, otherwise
        ``iter_get`` raises an error.
        """

        self.input_connector = MockTarget().input

    def test_raises_stop_iteration_on_kill_signal(self) -> None:
        """Test the iterator exits once it reaches a KillSignal object"""

        self.input_connector._queue.put(KillSignal)
        with self.assertRaises(StopIteration):
            next(self.input_connector.iter_get())

    def test_raises_missing_connection_with_no_parent(self) -> None:
        """Test the iterator exits if input has no paren"""

        with self.assertRaises(MissingConnectionError):
            next(Input().iter_get())

    def test_returns_queue_value(self) -> None:
        """Test the ``get`` method retrieves data from the underlying queue"""

        test_val = 'test_val'
        self.input_connector._queue.put(test_val)
        time.sleep(1)  # Give queue time to update

        self.assertEqual(next(self.input_connector.iter_get()), test_val)


class MaxSize(TestCase):
    """Tests the setting/getting of the maximum size for the underlying queue"""

    def setUp(self) -> None:
        self.connector = Input(maxsize=10)

    def test_queue_matches_input(self) -> None:
        """Test the max size of the underlying queue matches the max size of the Input"""

        self.assertEqual(self.connector._queue._maxsize, self.connector.maxsize)

    def test_set_at_init(self) -> None:
        """Test the max queue size is set at __init__"""

        self.assertEqual(10, self.connector.maxsize)

    def test_changed_via_setter(self) -> None:
        """Test the size of the underlying queue is changed when setting the ``maxsize`` attribute"""

        self.connector.maxsize = 5
        self.assertEqual(5, self.connector.maxsize)

    def test_error_on_nonempty_queue(self) -> None:
        """Test a ``RuntimeError`` is raised when changing the size of a nonempty connector"""

        self.connector._queue.put(1)
        time.sleep(1)  # Let the queue update

        with self.assertRaises(RuntimeError):
            self.connector.maxsize += 1


class QueueProperties(TestCase):
    """Test  test the exposure of queue properties by the overlying ``Connector`` class"""

    def setUp(self) -> None:
        self.connector = Input(maxsize=1)

    def test_size_matches_queue_size(self) -> None:
        """Test the ``size`` method returns the size of the queue`"""

        self.assertEqual(self.connector.size(), 0)
        self.connector._queue.put(1)
        self.assertEqual(self.connector.size(), 1)

    def test_full_state(self) -> None:
        """Test the ``full`` method returns the state of the queue"""

        self.assertFalse(self.connector.is_full())
        self.connector._queue.put(1)
        self.assertTrue(self.connector.is_full())

    def test_empty_state(self) -> None:
        """Test the ``empty`` method returns the state of the queue"""

        self.assertTrue(self.connector.is_empty())
        self.connector._queue.put(1)

        # The value of Queue.is_empty() updates asynchronously
        time.sleep(1)

        self.assertFalse(self.connector.is_empty())
